import atexit
import datetime
import logging
import os
import typing as t
from threading import Lock

from vbcore.uuid import get_uuid

try:
    from apscheduler import events
    from apscheduler.schedulers import SchedulerAlreadyRunningError
    from apscheduler.util import undefined
    from flask_apscheduler import APScheduler
except ImportError:  # pragma: no cover
    events = undefined = None
    SchedulerAlreadyRunningError = Exception
    BlockingScheduler = APScheduler = object

try:
    import fcntl

    THREAD_LOCK = Lock()
except ImportError:  # pragma: no cover
    THREAD_LOCK = fcntl = None  # type: ignore


class APJobs(APScheduler):
    def init_app(self, app):
        # this is necessary because super().init_app is conditionally invoked
        self.app = app

        if APScheduler is object:
            raise ImportError(
                "you must install 'flask_apscheduler'"
            )  # pragma: no cover

        self.set_config(app)

        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["scheduler"] = self

        if app.config.SCHEDULER_PATCH_MULTITHREAD is True:
            if fcntl is None:  # pragma: no cover
                app.logger.warning(
                    "fcntl not supported on this platform, no locking mechanism used"
                )
            elif not self.set_lock(app.config.SCHEDULER_LOCK_FILE):
                app.logger.info("scheduler already started in another process")
                return

        if not self.running and app.config.SCHEDULER_AUTO_START is True:
            try:
                super().init_app(app)
                self.add_listener(self.exception_listener, events.EVENT_ALL)
                self.start()
            except SchedulerAlreadyRunningError as exc:  # pylint: disable=broad-except
                app.logger.exception(exc)

    @staticmethod
    def set_config(app):
        app.config.setdefault("SCHEDULER_AUTO_START", False)
        app.config.setdefault("SCHEDULER_PATCH_MULTITHREAD", True)
        app.config.setdefault("SCHEDULER_LOCK_FILE", ".scheduler.lock")

        if app.debug:
            logger = logging.getLogger("apscheduler")
            logger.setLevel(logging.DEBUG)
            logger = logging.getLogger("flask_apscheduler")
            logger.setLevel(logging.DEBUG)

    # pylint: disable=too-many-arguments
    def add(
        self,
        func: t.Optional[t.Union[str, t.Callable]],
        trigger: t.Optional[t.Union[str, t.Callable]] = None,
        args: t.Optional[tuple] = None,
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        name: t.Optional[str] = None,
        misfire_grace_time: int = undefined,
        coalesce: bool = undefined,
        max_instances: int = undefined,
        next_run_time: t.Optional[datetime.datetime] = undefined,
        jobstore: str = "default",
        executor: str = "default",
        replace_existing: bool = False,
        **kw,
    ):
        if not self.running:
            self.app.logger.warning(
                "scheduler not running or not started in this process"
            )

        job = self.add_job(
            id=get_uuid(),
            func=func,
            trigger=trigger,
            args=args,
            kwargs=kwargs,
            name=name,
            misfire_grace_time=misfire_grace_time,
            coalesce=coalesce,
            max_instances=max_instances,
            next_run_time=next_run_time,
            jobstore=jobstore,
            executor=executor,
            replace_existing=replace_existing,
            **kw,
        )
        self.app.logger.debug("added job %s: %s", func, job.id)
        return job

    def set_lock(self, lock_file: str) -> bool:
        """
        ensure that only one worker starts the scheduler
        :return: (boolean) True -> scheduler can start
        """
        file = None

        def unlock():
            fcntl.flock(file, fcntl.LOCK_UN)
            file.close()
            os.remove(lock_file)

        try:
            with THREAD_LOCK:
                # pylint: disable=consider-using-with
                file = open(lock_file, "w", encoding="utf-8")
                file.write(str(os.getpid()))
                file.flush()
                fcntl.flock(file, fcntl.LOCK_EX | fcntl.LOCK_NB)
                atexit.register(unlock)
            return True
        except OSError as exc:
            self.app.logger.exception(exc)
            return False

    def exception_listener(self, event: "events.JobExecutionEvent"):
        code = event.code
        logger = self.app.logger
        if code == events.EVENT_JOB_ERROR:
            logger.error("An error occurred when executing job: %s", event.job_id)
            if isinstance(event, events.JobExecutionEvent):
                logger.exception(event.exception)
                logger.error(event.traceback)
        elif code == events.EVENT_JOB_ADDED:
            logger.info("successfully added job: %s", event.job_id)
        elif code == events.EVENT_JOB_EXECUTED:
            logger.debug(
                "successfully executed job: %s, returned value: %s",
                event.job_id,
                event.retval,
            )
        elif isinstance(event, events.JobEvent):
            logger.debug("received event (%s) for job: %s", event, event.job_id)
        else:
            logger.debug("received event: %s", event)



from flaskel.config import config

try:
    from flask_socketio import SocketIO

    IO_ASYNC_MODE = config("IO_ASYNC_MODE", default=None)
    IO_CORS_ORIGINS = config("IO_CORS_ORIGINS", default="*")
    IO_LOG_ENABLED = config("IO_LOG_ENABLED", default=True, cast=bool)

    socketio = SocketIO(
        async_mode=IO_ASYNC_MODE,
        logger=IO_LOG_ENABLED,
        engineio_logger=IO_LOG_ENABLED,
        cors_allowed_origins=IO_CORS_ORIGINS,
    )
except ImportError:
    socketio = None  # pylint: disable=invalid-name



import typing as t

import sqlalchemy as sa
from sqlalchemy.exc import SQLAlchemyError
from vbcore.datastruct import ObjectDict
from vbcore.db.mixins import StandardMixin
from vbcore.jsonschema.support import Fields

from flaskel import cap, db_session

from .exceptions import MediaError

SCHEMA_MEDIA = Fields.array_object(
    properties={
        "id": Fields.integer,
        "link": Fields.string,
    }
)


class MediaMixin(StandardMixin):
    link = sa.Column(sa.String(255), nullable=False, unique=True)
    path = sa.Column(sa.String(255), nullable=False, unique=True)
    filename = sa.Column(sa.String(255))

    def to_dict(self, *_, **__):
        return ObjectDict(id=self.id, link=self.link)


class MediaRepo:
    entity_model = None
    media_model = None
    session = db_session

    @classmethod
    def entity_name(cls) -> t.Optional[str]:
        if cls.entity_model:
            return cls.entity_model.__tablename__
        return None

    @classmethod
    def get(cls, eid):
        return cls.session.query(cls.entity_model).get_or_404(eid)

    @classmethod
    def update(cls, emodel, **kwargs):
        res = cls.media_model(**kwargs)  # pylint: disable=not-callable
        emodel.images.append(res)
        return res

    @classmethod
    def store(cls, emodel):
        try:
            cls.session.add(emodel)
            cls.session.commit()
        except SQLAlchemyError as exc:
            cap.logger.exception(exc)
            cls.session.rollback()
            raise MediaError(exc) from exc

    @classmethod
    def delete(cls, entity_id, media_id):
        try:
            entity = getattr(cls.media_model, cls.entity_name())

            res = (
                cls.session.query(cls.media_model)
                .filter(entity.any(id=entity_id))
                .filter_by(id=media_id)
                .first_or_404()
            )
            cls.session.delete(res)
            cls.session.commit()
            return res
        except SQLAlchemyError as exc:
            cap.logger.exception(exc)
            cls.session.rollback()
            raise MediaError(exc) from exc



import logging
import typing as t
from datetime import datetime

import flask
from packaging import version
from vbcore.http import httpcode, HttpMethod
from vbcore.http.headers import ContentTypeEnum, HeaderEnum

from flaskel import abort, cap, ExtProxy, Flaskel, request, Response, webargs
from flaskel.ext.default import builder
from flaskel.ext.limit import RateLimit
from flaskel.views import BaseView


class RedisStore:
    def __init__(self, redis=None, sep: str = "/"):
        self.sep = sep
        self.client = redis

    @staticmethod
    def normalize_string(data) -> t.Optional[str]:
        if not data:
            return None
        return data if isinstance(data, str) else data.decode()

    def _normalize(self, ver: str):
        s = ver.split(self.sep)
        return version.parse(s[0]), bool(int(s[1])) if len(s) > 1 else False

    def release(self, key: str, v: str, u: bool = False):
        if self.client.lpush(key, f"{v}{self.sep}{int(u)}"):
            return version.parse(v), u
        return None

    def latest(self, key: str):
        ver = self.retrieve(key)
        if len(ver):
            return ver[0]
        return None, None

    def pop(self, key: str):
        res = self.normalize_string(self.client.lpop(key))
        return self._normalize(res) if res else None

    def clear(self, key: str):
        return self.client.delete(key)

    def retrieve(self, key: str, max_item: int = 1):
        data = self.client.lrange(key, 0, max_item - 1)
        return [self._normalize(self.normalize_string(d)) for d in data]


class MobileVersionCompatibility:
    def __init__(
        self, app: t.Optional[Flaskel] = None, store=None, current_version=None
    ):
        self._store = store
        self._load_time: t.Optional[datetime] = None
        self._current_version = current_version
        self._versions: t.Dict[t.Optional[str], list] = {None: []}

        if app is not None:
            self.init_app(app, store, current_version)

    def init_app(self, app: Flaskel, store=None, current_version=None):
        self._store = store or self._store
        self._current_version = current_version or self._current_version
        if not self._current_version:
            self._current_version = app.version

        self.load_config(app)

        if app.config.VERSION_CHECK_ENABLED:
            try:
                self.load_from_storage()
            except Exception as exc:  # pylint: disable=broad-except
                app.logger.exception(exc)

            app.before_request_funcs.setdefault(None, []).append(
                self._set_mobile_version
            )
            app.after_request_funcs.setdefault(None, []).append(
                self._set_version_headers
            )

        app.extensions["mobile_version"] = self

    @property
    def mobile_version(self):
        return flask.g.get("mobile_version")

    @property
    def store(self):
        return self._store

    @staticmethod
    def load_config(app):
        app.config.setdefault("VERSION_STORE_MAX", 6)
        app.config.setdefault("VERSION_CACHE_EXPIRE", 3600)
        app.config.setdefault("VERSION_CHECK_ENABLED", True)
        app.config.setdefault("VERSION_AGENT_HEADER", "X-Agent")
        app.config.setdefault("VERSION_API_HEADER", "X-Api-Version")
        app.config.setdefault("VERSION_STORE_KEY", "x_upgrade_needed")
        app.config.setdefault("VERSION_HEADER_KEY", "X-Mobile-Version")
        app.config.setdefault("VERSION_UPGRADE_HEADER", "X-Upgrade-Needed")
        app.config.setdefault("VERSION_AGENTS", ("android", "ios"))
        app.config.setdefault(
            "VERSION_SKIP_STATUSES",
            (
                httpcode.FORBIDDEN,
                httpcode.NOT_FOUND,
                httpcode.METHOD_NOT_ALLOWED,
                httpcode.TOO_MANY_REQUESTS,
            ),
        )

    @staticmethod
    def version_parse(ver: str) -> version.Version:
        return version.Version(ver)

    @staticmethod
    def _set_mobile_version():
        # pylint: disable=assigning-non-slot
        flask.g.mobile_version = request.headers.get(cap.config.VERSION_HEADER_KEY)

    @staticmethod
    def agent_identity() -> t.Optional[str]:
        family = flask.g.user_agent.os.family
        return family.lower() if family else None

    def _get_agent(self) -> t.Optional[str]:
        agent = request.headers.get(cap.config.VERSION_AGENT_HEADER)
        if agent in cap.config.VERSION_AGENTS:
            return agent
        flask.g.user_agent.parse()
        return self.agent_identity()

    def _set_version_headers(self, resp):
        if resp.status_code not in cap.config.VERSION_SKIP_STATUSES:
            resp.headers[cap.config.VERSION_UPGRADE_HEADER] = int(self.check_upgrade())
            resp.headers[cap.config.VERSION_API_HEADER] = self._current_version
        return resp

    def _agent_key(self, agent: str) -> str:
        return f"{cap.config.VERSION_STORE_KEY}{self._store.sep}{agent}"

    def latest(self, agent: str) -> str:
        versions = self._versions.get(agent) or []
        return str(versions[0][0]) if len(versions) > 0 else ""

    def all_releases(self, agent: str) -> t.List[t.Dict[str, t.Any]]:
        versions = self._versions.get(agent) or []
        return [{"version": str(v), "critical": u} for v, u in versions]

    def rollback(self, agent: str):
        self._store.pop(self._agent_key(agent))
        if agent in self._versions and len(self._versions[agent]):
            self._versions[agent] = self._versions[agent][1:]

    def clear(self, agent: str):
        self._store.clear(self._agent_key(agent))
        if agent in self._versions:
            self._versions[agent] = []

    def publish(self, agent: str, ver: str, critical: bool = False):
        latest, _ = self._store.latest(self._agent_key(agent))
        if latest is not None and latest >= self.version_parse(ver):
            raise ValueError("New version must be greater than latest")

        self._store.release(self._agent_key(agent), ver, critical)
        self.load_from_storage(force=True)
        return self._versions[agent]

    def load_from_storage(self, force: bool = False):
        if self._load_time:
            seconds = (datetime.now() - self._load_time).seconds
            if not force and seconds < cap.config.VERSION_CACHE_EXPIRE:
                return
        for a in cap.config.VERSION_AGENTS:
            smax = cap.config.VERSION_STORE_MAX
            self._versions[a] = self._store.retrieve(self._agent_key(a), smax)
        self._load_time = datetime.now()

    def check_upgrade(self) -> bool:
        if not self.mobile_version:
            return False
        try:
            self.load_from_storage()
            agent = self._get_agent()
            versions = self._versions.get(agent) or []
            mv = self.version_parse(self.mobile_version)
            for v, u in versions:
                if v > mv and u is True:
                    return True
                if v <= mv:
                    return False
        except Exception as exc:  # pylint: disable=broad-except
            cap.logger.exception(exc)
            return False
        return len(versions) >= cap.config.VERSION_STORE_MAX


class MobileReleaseView(BaseView):
    builder = builder
    ext: MobileVersionCompatibility = t.cast(
        MobileVersionCompatibility, ExtProxy("mobile_version")
    )
    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [
        HttpMethod.POST,
        HttpMethod.GET,
        HttpMethod.DELETE,
    ]

    default_view_name = "mobile_release"
    default_urls = (
        "/mobile/release",
        "/mobile/release/<ver>",
    )

    @webargs.query(
        dict(
            all=webargs.OptField.boolean(),
            critical=webargs.OptField.boolean(),
            agent=webargs.Field.string(required=True),
        )
    )
    def dispatch_request(self, params: dict, *_, ver=None, **__):
        agent = params["agent"]
        if agent not in cap.config.VERSION_AGENTS:
            return abort(
                httpcode.BAD_REQUEST,
                response=Response(
                    {
                        "reason": "agent not compatible",
                        "agents": cap.config.VERSION_AGENTS,
                    }
                ),
            )

        if ver == "latest":
            return self.ext.latest(agent) or "", {
                HeaderEnum.CONTENT_TYPE: ContentTypeEnum.PLAIN
            }

        if ver is None:
            if request.method == "DELETE":
                if params["all"]:
                    self.ext.clear(agent)
                    return Response.no_content()
                self.ext.rollback(agent)
        else:
            try:
                self.ext.publish(agent, ver, params["critical"])
            except ValueError as exc:
                abort(httpcode.BAD_REQUEST, str(exc))

        mimetype, response = self.builder.get_mimetype_accept()
        return response.build(self.ext.all_releases(agent)), {
            HeaderEnum.CONTENT_TYPE: mimetype
        }


class MobileLoggerView(BaseView):
    unavailable = "N/A"
    intro = "An exception occurred on mobile app:"
    default_view_name = "mobile_logger"
    default_urls = ("/mobile/logger",)

    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [
        HttpMethod.POST,
    ]
    decorators = [
        builder.no_content,
        RateLimit.medium(),
        RateLimit.fail(),
    ]

    def __init__(self, name: t.Optional[str] = None):
        self._log = logging.getLogger(name) if name else cap.logger

    @staticmethod
    def _dump_dict(data: dict) -> str:
        return "\n\t".join(f"{k}: {v}" for k, v in data.items())

    def dump_key(self, data: dict, key: str) -> t.Optional[str]:
        if key not in data:
            return self.unavailable

        _data = data.get(key)
        if isinstance(_data, dict):
            return self._dump_dict(_data)
        return _data

    def get_user_info(self, *_, **__) -> str:
        return self.unavailable

    def dump_message(self, payload: dict, *args, **kwargs) -> t.Tuple[str, tuple]:
        return (
            (
                "%s"
                "\n\tUser-Info: %s"
                "\n\tMobile-Version: %s"
                "\n\tUser-Agent: %s"
                "\n\tDevice-Info: %s"
                "\n\tDebug-Info: %s"
                "\n\tPayload: %s"
                "\n\tStack-Trace:\n%s"
            ),
            (
                self.intro,
                self.get_user_info(*args, **kwargs),
                self.dump_key(request.headers, "X-Mobile-Version"),
                self.dump_key(request.headers, "User-Agent"),
                self.dump_key(payload, "device_info"),
                self.dump_key(payload, "debug_info"),
                self.dump_key(payload, "payload"),
                self.dump_key(payload, "stacktrace").replace("\\n", "\n"),
            ),
        )

    def perform(self, payload: dict, *args, **kwargs) -> t.Tuple[str, tuple]:
        message, args = self.dump_message(payload, *args, **kwargs)
        self._log.error(message, *args)
        return message, args

    def dispatch_request(self, *args, **kwargs):
        payload = request.json
        self.perform(payload, *args, **kwargs)

        if "debug" in request.args:
            return payload, httpcode.SUCCESS
        return None



import typing as t
from functools import partial

from bson import ObjectId
from flask_pymongo import PyMongo
from flask_pymongo.wrappers import Collection
from pymongo.cursor import Cursor
from pymongo.results import DeleteResult
from vbcore.datastruct import ObjectDict
from vbcore.date_helper import Seconds
from vbcore.http import httpcode

from flaskel import client_mongo, ConfigProxy, Response
from flaskel.utils.datastruct import Pagination


class FlaskMongoDB(PyMongo):
    def init_app(self, app, *args, ext_name: str = "default", **kwargs):
        self.set_default_config(app, **kwargs)
        super().init_app(app, *args, **app.config.MONGO_OPTS)
        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["mongo"] = ObjectDict()
        app.extensions["mongo"][ext_name] = self

    @classmethod
    def set_default_config(cls, app, **kwargs):
        app.config.setdefault("MONGO_URI", "mongodb://localhost")
        app.config.setdefault("MONGO_OPTS", {})

        timeout = app.config.MONGO_OPTS.pop("timeout", None)
        default_timeout = timeout or Seconds.millis

        app.config.MONGO_OPTS.setdefault("appname", app.config.APP_NAME)
        app.config.MONGO_OPTS.setdefault("directConnection", False)
        app.config.MONGO_OPTS.setdefault("minPoolSize", 0)
        app.config.MONGO_OPTS.setdefault("maxPoolSize", 100)
        app.config.MONGO_OPTS.setdefault("maxConnecting", 2)
        app.config.MONGO_OPTS.setdefault("retryWrites", True)
        app.config.MONGO_OPTS.setdefault("retryReads", True)
        app.config.MONGO_OPTS.setdefault("maxIdleTimeMS", None)
        app.config.MONGO_OPTS.setdefault("waitQueueTimeoutMS", None)
        app.config.MONGO_OPTS.setdefault("document_class", ObjectDict)
        app.config.MONGO_OPTS.setdefault("socketTimeoutMS", default_timeout)
        app.config.MONGO_OPTS.setdefault("connectTimeoutMS", default_timeout)
        app.config.MONGO_OPTS.setdefault("serverSelectionTimeoutMS", default_timeout)
        app.config.MONGO_OPTS.setdefault("heartbeatFrequencyMS", Seconds.millis * 10)

        app.config.MONGO_OPTS.update(**kwargs)


ResIdType = t.Union[ObjectId, str]
SortType = t.List[t.Tuple[str, int]]


class BaseRepo:
    sink = client_mongo
    collection_key: str = ""
    collections = ConfigProxy("COLLECTIONS")

    sort_by: t.Optional[SortType] = None
    projection_list: t.Optional[t.List[str]] = None
    projection_detail: t.Optional[t.List[str]] = None

    @classmethod
    def prepare_record(cls, record: dict):
        return record

    @classmethod
    def prepare_collection(cls, collection: t.Optional[str] = None) -> str:
        return cls.collections.get(collection or cls.collection_key)

    @classmethod
    def connection(cls, collection: t.Optional[str] = None) -> Collection:
        return cls.sink[cls.prepare_collection(collection)]

    @classmethod
    def range_filter(cls, flt_from, flt_to) -> dict:
        return {"$gte": flt_from, "$lte": flt_to}

    @classmethod
    def count(
        cls, collection: t.Optional[str] = None, filters: t.Optional[dict] = None
    ) -> int:
        return cls.connection(collection).count_documents(filters or {})

    @classmethod
    def aggregate(
        cls, stages: t.List[dict], collection: t.Optional[str] = None, **kwargs
    ):
        return cls.connection(collection).aggregate(stages, **kwargs)

    @classmethod
    def find(
        cls,
        collection: t.Optional[str] = None,
        filters: t.Optional[dict] = None,
        projection: t.Optional[t.List[str]] = None,
        sort: t.Optional[SortType] = None,
        raise_404: bool = False,
        **kwargs,
    ) -> Cursor:
        conn = cls.connection(collection)
        find = conn.find_one_or_404 if raise_404 else conn.find
        return find(filters or {}, projection, sort=sort, **kwargs)

    @classmethod
    def delete(
        cls,
        res_id: ResIdType,
        collection: t.Optional[str] = None,
        **kwargs,
    ) -> DeleteResult:
        return cls.connection(collection).delete_one(
            {"_id": ObjectId(res_id) if isinstance(res_id, str) else res_id, **kwargs}
        )

    @classmethod
    def get_list(
        cls,
        filters: t.Optional[dict] = None,
        pagination: t.Optional[Pagination] = None,
        projection: t.Optional[t.List[str]] = None,
        sort: t.Optional[SortType] = None,
        collection: t.Optional[str] = None,
        **kwargs,
    ) -> t.Union[t.Tuple[list, int, dict], t.List[dict]]:
        cursor = partial(
            cls.find,
            collection=collection,
            filters=filters,
            projection=projection or cls.projection_list,
            sort=sort or cls.sort_by,
            **kwargs,
        )

        if pagination:
            total = cls.count(collection=collection, filters=filters)
            query = cursor() if total else ()
            _cursor = (
                query.skip(pagination.offset()).limit(pagination.per_page())  # type: ignore
                if query != ()
                else tuple()
            )
            return (
                [cls.prepare_record(d) for d in _cursor],
                (
                    httpcode.PARTIAL_CONTENT
                    if pagination.is_paginated(total)
                    else httpcode.SUCCESS
                ),
                Response.pagination_headers(total, pagination),
            )

        return [cls.prepare_record(d) for d in cursor()]

    @classmethod
    def get_detail(
        cls,
        res_id: ResIdType,
        filters: t.Optional[dict] = None,
        projection: t.Optional[t.List[str]] = None,
        sort: t.Optional[SortType] = None,
        **kwargs,
    ):
        return cls.find(
            raise_404=True,
            filters={
                "_id": ObjectId(res_id) if isinstance(res_id, str) else res_id,
                **(filters or {}),
            },
            projection=projection or cls.projection_detail,
            sort=sort or cls.sort_by,
            **kwargs,
        )
